use std::error::Error;

use greetd_ipc::{codec::SyncCodec, AuthMessageType, ErrorType, Request, Response};

use crate::{AuthStatus, Greeter, Mode};

pub fn handle(greeter: &mut Greeter) -> Result<(), Box<dyn Error>> {
  if let Some(ref request) = greeter.request {
    request.write_to(&mut greeter.stream())?;
    greeter.request = None;
    let response = Response::read_from(&mut greeter.stream())?;

    parse_response(greeter, response)?;
  }

  Ok(())
}

fn parse_response(greeter: &mut Greeter, response: Response) -> Result<(), Box<dyn Error>> {
  match response {
    Response::AuthMessage { auth_message_type, auth_message } => match auth_message_type {
      AuthMessageType::Secret => {
        greeter.mode = Mode::Password;
        greeter.working = false;
        greeter.secret = true;
        greeter.prompt = auth_message;
      }

      AuthMessageType::Visible => {
        greeter.mode = Mode::Password;
        greeter.working = false;
        greeter.secret = false;
        greeter.prompt = auth_message;
      }

      AuthMessageType::Error => greeter.message = Some(auth_message),

      AuthMessageType::Info => {
        if let Some(message) = &mut greeter.message {
          message.push_str(&auth_message.trim_end());
          message.push('\n');
        } else {
          greeter.message = Some(auth_message.trim_end().to_string());

          if let Some(message) = &mut greeter.message {
            message.push('\n');
          }
        }

        Request::PostAuthMessageResponse { response: None }.write_to(&mut greeter.stream())?;
        greeter.request = None;
        let response = Response::read_from(&mut greeter.stream())?;

        parse_response(greeter, response)?;
      }
    },

    Response::Success => {
      if greeter.done {
        crate::exit(greeter, AuthStatus::Success)?;
      } else if let Some(command) = &greeter.command {
        greeter.done = true;
        greeter.request = Some(Request::StartSession { cmd: vec![command.clone()] });
      } else {
        crate::exit(greeter, AuthStatus::Failure)?;
      }
    }

    Response::Error { error_type, description } => match error_type {
      ErrorType::AuthError => crate::exit(greeter, AuthStatus::Failure)?,

      ErrorType::Error => {
        cancel(greeter);
        greeter.message = Some(description)
      }
    },
  }

  Ok(())
}

pub fn cancel(greeter: &mut Greeter) {
  Request::CancelSession.write_to(&mut greeter.stream()).unwrap();
}
